#################################
# This file was created by Alessandro Conigli - 2022
# The pourpose of this code is to compute twisted mass 
# observables from raw correlators stored in mesons.dat files.
# The analysis is performed at the level of a single ensemble 
# and results are then stored in BDIO files 
# It extracts:
#  - charm quark mass 
#  - ll hl hh pseudo and vector masses 
#  - ll hl hh pseudo and vector leptonic decays
#  
#  The analysis is performed by solving a GEVP 
##################################
using Base: String
using Base: @kwdef
using LaTeXStrings: length
using OrderedCollections
using juobs, BDIO, DelimitedFiles, ADerrors, LaTeXStrings, PyPlot

#============= SET UP VARIABLES ===========#
rcParams = PyPlot.PyDict(PyPlot.matplotlib."rcParams")
rcParams["text.usetex"] =  false
rcParams["mathtext.fontset"]  = "cm"
rcParams["font.size"] =10
rcParams["axes.labelsize"] =22
rcParams["axes.titlesize"] = 18

plt.rc("text", usetex=false) # set to true once you install latex

#============ PATHS & INFO =================#
# path to data 
const path_data_w  = "/Users/alessandroconigli/Lattice/data/aux_obs_data/wilson"
const path_data_tm = "/Users/alessandroconigli/Lattice/data/charm_full_Dirac"
# path to plateau
const path_plat_w       = "/Users/alessandroconigli/Lattice/data/aux_obs_data/plat_wilson.txt"
const path_plat_tm_mps  = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_mps.txt"
const path_plat_tm_mvec = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_mvec.txt"
const path_plat_tm_dps  = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_dps.txt"
const path_plat_tm_dvec = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_dvec.txt"
# path to aux obs 
const path_rw = "/Users/alessandroconigli/Lattice/data/aux_obs_data/rwf"
const path_md = "/Users/alessandroconigli/Lattice/data/aux_obs_data/md"
# path to bdio
const path_bdio_md         = "/Users/alessandroconigli/Lattice/data/bdio_charm/md"
const path_bdio_tm         = "/Users/alessandroconigli/Lattice/data/bdio_charm/tm"
const path_bdio_tm_shifted = "/Users/alessandroconigli/Lattice/data/bdio_charm/tm_shifted"
const path_plot = "/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/plots"

# ensembles to analyse 
const ens_list = ["D200"]
# reweighting and mass shift flags
const rwf = true
const mass_shift = true 
const tau = 4    # gevp shift parameter
const TSM = false # set whether TSM is used or not 

#@warning("\nTSM FLAG MODE: $(TSM)\n")

#=============== INCLUDES ===============#

include("types.jl")
include("tools.jl")
include("const.jl")
include("read_bdio.jl")

#=============== ENSEMBLE INFO FROM DATABASE ==================#
ensinfo = Vector{EnsInfo}(undef, length(ens_list))
for i in 1:length(ens_list)
    ens = ens_list[i]
    try
        ensinfo[i]= EnsInfo(ens, ens_db[ens], trunc_db[ens])
    catch
        error("The ensemble id ", ens, " was not found in the const.jl ens_db database.
        Please check the ensemble id or update the database")
    end
end
wpmm = Dict{String, Vector{Float64}}()
wpmm["H101"] = [-1.0, 4.0, -1.0, 14.0*2.86]
wpmm["H102r002"] = [-1.0, 4.0, -1.0, 14.0*2.86]
wpmm["H105"] = [-1.0, 4.0, -1.0, 14.0*2.86]
wpmm["H400"] = [-1.0, 4.0, -1.0, 14.0*3.659]
wpmm["N202"] = [1.0, -2.0, -1.0, -1.0] #14.0*5.164]
wpmm["N200"] = [-1.0, 4.0, -1.0, 14.0*5.164]
wpmm["N203"] = [-1.0, 4.0, -1.0, 14.0*5.164]
#wpmm["D200"] = [-1.0, 4.0, -1.0, 14.0*5.164]
wpmm["D200"] = [1.0, -1.0, -1.0, -1.0]
wpmm["N300"] = [-1.0, 4.0, -1.0, 14.0*8.595]
wpmm["J303"] = [-1.0, 4.0, -1.0, 14.0*8.595]
wpmm["D200_corr"] = [1.0, -1.0, -1.0, -1.0]

##

##
#=============== ANALYSIS ===============#

gen_mulist = Vector{Vector{Vector{Float64}}}(undef, length(ens_list))

# obs storage  
obs_tm = Vector(undef, length(ensinfo))
if mass_shift
    deltam = Vector(undef, length(ensinfo))
    obs_tm_shifted = Vector(undef, length(ensinfo))
end

pp_tm= Vector(undef, length(ensinfo))
pp_tmW= Vector(undef, length(ensinfo))
a0a0_tm= Vector(undef, length(ensinfo))
pa0_tm= Vector(undef, length(ensinfo))
a1a1_tm= Vector(undef, length(ensinfo))
pa1_tm= Vector(undef, length(ensinfo))
for (k,ens) in enumerate(ensinfo)
    #t0_ens, YW, W = get_t0(path_data_w, ens, rw=rwf, path_rw=path_rw)
    pp_tm[k], pp_tmW[k] = get_corr(path_data_tm, ens,"G5", "G5", rw=rwf, path_rw=path_rw, tsm=TSM, ex=true)
    
    a0a0_tm[k], _    = get_corr(path_data_tm, ens, "G0G5", "G0G5", rw=rwf, path_rw=path_rw, tsm=TSM, ex=true)
    a1a1_tm[k], _    = get_corr(path_data_tm, ens, "G1G5", "G1G5", rw=rwf, path_rw=path_rw, tsm=TSM, ex=true)
    # a2a2_tm, _    = get_corr(path_data_tm, ens, "G2G5", "G2G5", rw=rwf, path_rw=path_rw, tsm=TSM)
    # a3a3_tm, _    = get_corr(path_data_tm, ens, "G3G5", "G3G5", rw=rwf, path_rw=path_rw, tsm=TSM)

    pa0_tm[k], _     = get_corr(path_data_tm, ens, "G5", "G0G5", rw=rwf, path_rw=path_rw, tsm=TSM, ex=true)
    pa1_tm[k], _     = get_corr(path_data_tm, ens, "G5", "G1G5", rw=rwf, path_rw=path_rw, tsm=TSM, ex=true)
    # pa2_tm, _     = get_corr(path_data_tm, ens, "G5", "G2G5", rw=rwf, path_rw=path_rw, tsm=TSM)
    # pa3_tm, _     = get_corr(path_data_tm, ens, "G5", "G3G5", rw=rwf, path_rw=path_rw, tsm=TSM)
    
end
##

mm = Vector{Vector{uwreal}}(undef, length(ensinfo))
for (k,ens) in enumerate(ensinfo)
    mu_list = gen_mulist[k] = getfield.(pp_tm[k], :mu)
    # gevp matrices mixing different gamma struct 
    # mat_pp = comp_mat_multigamma(ens, pp_tm[k], pa1_tm[k], a1a1_tm[k])[1:9]
    
    # gevp matrices shifting in time 
    mat_pp = comp_mat_tau_shift(ens, pp_tm[k], tau)

    # mps0, mps1 =  gevp_mass(mat_pp, path_plat_tm_mps, path_plat_tm_mvec, t0=3, wpm=wpmm, pl=true)
    mps0  =  gevp_mass_BMA(mat_pp, t0=2, wpm=wpmm, pl=true)
    mm[k] =  mps0
end


##
mefff = meff(pp_tm[1][1], [90,95], pl=true, wpm=wpmm)

##
""" this function return a vector of correlators built using gevp projection to the ground state. 
A GEVP is solved for several ta and tb and a new correlator is computed for each of them
"""
function gevp_GPOF(ensinfo::EnsInfo, corr::Vector{juobs.Corr}; N::Int64=4, TA::Vector{Int64}=[2,4], TB::Vector{Int64}=[7,8])
    mat_info = comp_mat_GPOF(ensinfo, corr, N=N)
    y0 = getfield(corr[1], :y0)
    mu = getfield(corr[1], :mu)
    new_corr = Vector{Vector{Vector{uwreal}}}(undef, length(mat_info))

    for i in eachindex(mat_info)
        mat = getfield(mat_info[i], :mat_list)
        new_corr[i] = Vector{Vector{uwreal}}(undef, 0)
        for ta in TA
            for tb in TB
                evalues, evecs = juobs.uweigen(mat[ta], mat[tb], iter = 20)
                aux_corr = [uwdot(juobs.transpose(evecs[:,N]), uwdot( mat[k], evecs[:,N]) )[1] for k in eachindex(mat)]
                push!(new_corr[i], aux_corr)
            end 
        end 
    end
    return new_corr
end

function gevp_GPOF_BMA(ensinfo::EnsInfo, corr::Vector{juobs.Corr}; N::Int64=4, TA::Vector{Int64}=[2], TB::Vector{Int64}=[7,8], pl::Bool=true, wpm::Union{Nothing, Dict{String,Vector{Float64}}}=nothing)
    new_corr = gevp_GPOF(ensinfo, corr, N=N, TA=TA, TB=TB)
    @. one_exp_fit(x,p) = p[1] + p[2]*exp(-p[3]*x)
    Nparam=3
    y0 = getfield(corr[1], :y0)
    T = length(corr[1].obs) - y0
    #  fit time ranges for bayesian_av
    tmin = collect(10:12)
    tmax = collect(T-40:T-39)

    MAve = Vector{uwreal}(undef, length(corr))


    for k in eachindex(new_corr) # mass index 
        AIC = Vector{Float64}(undef, 0)
        Masses = Vector{uwreal}(undef, 0)
        for j in eachindex(new_corr[k]) # ta and tb gevp index 
            _, data_meff = juobs.meff(new_corr[k][j], [tmin[end], tmax[1]], pl=true, data=true, wpm=wpm)
            _, _, param, _, aic = juobs.bayesian_av(one_exp_fit, data_meff, tmin, tmax, Nparam, pl=false, wpm=wpmm, data = true)
            close("all")
            AIC  = vcat(AIC, aic )
            Masses  = vcat(Masses, param)
        end
        isnothing(wpm) ? uwerr.(Masses) : [uwerr(Masses[i], wpm) for i in eachindex(Masses)]
        offset = minimum(AIC)
        AIC = AIC .- offset
        Wtot = exp.(-0.5 .* AIC)
        Wtot /= sum(Wtot)
        MAve[k] = sum(Masses .* Wtot) 
        isnothing(wpm) ? uwerr(MAve[k]) : uwerr(MAve[k], wpm)

        if pl
            
            subplot(211)
            title(string(L"$\mu_1 = $", corr[k]. mu[1], L" $\mu_2 = $", corr[k]. mu[2]))
            ax1 = gca()
            setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
            xx = collect(1:length(Masses))
            vv = value.(Masses)
            errorbar(xx, vv, err.(Masses), fmt="s")
            aux1 = sum( Wtot .* Masses.^2)
            aux2 = sum( Wtot .* Masses)^2
            syst = sqrt(abs(value(aux1-aux2)))
            fill_between(xx, value(MAve[k]) - err(MAve[k]), value(MAve[k]) + err(MAve[k]),color="red", alpha=0.2)
            ylabel(L"$M_0$")

            subplot(212)
            plot(xx, Wtot)
            ylabel(L"$W$")
            xlabel("Models")
            display(gcf())
            close()
        end
    end
    return MAve
end
##  test gevp_mass_GPOF

mm_gpof = gevp_GPOF_BMA(ensinfo[1],pp_tm[1][9:9], pl=true, TA=[4], TB=[6], wpm=wpmm);
uwerr.(mm_gpof)
println(mm_gpof)
##
new_corr = gevp_GPOF(ensinfo[1], pp_tm[1][3:3], TA=[4], TB=[10]);
mtest = juobs.meff.(new_corr[3], fill([25,45], 4), pl=true, wpm=wpmm  ) 
mstd = meff(pp_tm[1][2], [65,80], pl=true, wpm=wpmm   ) 


##


function gevp_mass_BMA(mat_obs::Vector{MatInfo}; t0::Int64=2,pl::Bool=true, wpm::Union{Nothing, Dict{String,Vector{Float64}}}=nothing)
    y0s     = getfield.(mat_obs, :y0) 
    mu_list = getfield.(mat_obs, :mu) 
    en0     = Vector{uwreal}(undef, length(mat_obs))
    en1     = Vector{uwreal}(undef, length(mat_obs))
    data_tot = Vector(undef, length(mat_obs))
    W_tot = Vector(undef, length(mat_obs))

    for k in eachindex(mat_obs)
        mat = getfield(mat_obs[k], :mat_list) #[y0s[i]+1:end-1] #matrices to diagonalise for a given sector [i]
        evals = juobs.getall_eigvals(mat, t0)
        entot = juobs.energies(evals, wpm=wpm)

        @. gs_model(x,p) = p[1] + p[2] * exp(-(p[3])*x)
        param = 3
        tmin  =  collect(10:12) # scart from source 
        tmax  =  collect(36:38) 
        
        en0[k], _, data_tot[k], W_tot[k], _ = juobs.bayesian_av(gs_model, entot[1], tmin, tmax, param, pl=false, wpm=wpmm, data=true)
       close("all")
       if pl
            Masses = data_tot[k]
            isnothing(wpm) ? uwerr.(Masses) : [uwerr(Masses[i]) for i in eachindex(Masses)]
            subplot(211)
            # title(string(L"$\mu_1 = $", corr[k]. mu[1], L" $\mu_2 = $", corr[k]. mu[2]))
            ax1 = gca()
            setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
            xx = collect(1:length(Masses))
            vv = value.(Masses)
            errorbar(xx, vv, err.(Masses), fmt="s")
            aux1 = sum( W_tot[k] .* Masses.^2)
            aux2 = sum( W_tot[k] .* Masses)^2
            syst = sqrt(abs(value(aux1-aux2)))
            fill_between(xx, value(en0[k]) - err(en0[k]), value(en0[k]) + err(en0[k]), color="red", alpha=0.2)
            ylabel(L"$M_0$")

            subplot(212)
            plot(xx, W_tot[k])
            ylabel(L"$W$")
            xlabel("Models")
            display(gcf())
            close()
        end
        if pl 
            vv = value.(entot[1])
            ee = err.(entot[1])
            xx = collect(1:length(vv))
            errorbar(xx, vv, ee, fmt="*")
            # xlim(1,y0s[k]-10)
            ylim(value(en0[k])-10*err(en0[k]), value(en0[k])+10*err(en0[k]))
            display(gcf())
            close()
        end
    end
    return en0, data_tot, W_tot
end



## test uweigen

aa = comp_mat_GPOF(ensinfo[1], pp_tm[1])

mat1 = aa[9].mat_list[3]
mat2 = aa[9].mat_list[8]

ee, vv = juobs.uweigen(mat1, mat2, iter=10)

ee1, vv1 = eigen(value.(mat1), value.(mat2))
##
for k in 1:9
    println("\nk=$(k)")
    println("tau=4 ", mps0[k])
    println("tau=3 ", mps0_old_tau[k])

end
##
meffpp = juobs.meff(pp_tm[1][7], [85,90], pl=true)