#################################
# This file was created by Alessandro Conigli - 2022
# The pourpose of this code is to read BDIO files 
# and perform chiral+continuum extrapolation.
#  
##################################

using Base: String
using Base: @kwdef
using LaTeXStrings: length
using OrderedCollections, ProgressMeter
using juobs, BDIO, DelimitedFiles, ADerrors, LaTeXStrings, PyPlot, Statistics

#============= SET UP VARIABLES ===========#
rcParams = PyPlot.PyDict(PyPlot.matplotlib."rcParams")
rcParams["text.usetex"] =  true
rcParams["mathtext.fontset"]  = "cm"
rcParams["font.size"] =12
rcParams["axes.labelsize"] =20
rcParams["axes.titlesize"] = 18

plt.rc("text", usetex=true) # set to true once you install latex
#rcParams.update(PyPlot.matplotlib.mpl.rcParamsDefault)
#================== PATHS ==================#
# data by ale (carfull with the read_BDIO that you include)

# path to data 
const path_data_w  = "/Users/alessandroconigli/Lattice/data/aux_obs_data/wilson" # required for t0 at finite a
const path_rw = "/Users/alessandroconigli/Lattice/data/aux_obs_data/rwf"  # reweighting factors 
# path to bdio 
const path_bdio = "/Users/alessandroconigli/Lattice/data/bdio_charm/LAT22/" 
const path_bdio_tm = joinpath(path_bdio, "tm_shifted")
# const path_bdio_md = joinpath(path_bdio, "md") # path to aux obs (md) comptued by me 
const path_bdio_md = "/Users/alessandroconigli/Lattice/data/bdio_charm/AS_ligth_obs/" # path to delta m computed by AS
const path_bdio_t0 = "/Users/alessandroconigli/Lattice/data/bdio_charm/AS_ligth_obs/t0_combined.bdio" # path to t0_ph computed by AS
# path to t0 plateau
const path_plat_w       = "/Users/alessandroconigli/Lattice/data/aux_obs_data/plat_wilson.txt"


# data by javier (carful with the read_BDIO that you include. do not use the one in this folder)
# const path_bdio = "/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/analysis_with_javier_data/bdio" #PATH to BDIO folder
# const path_bdio_tm = joinpath(path_bdio, "tm_shifted")
# const path_bdio_md = joinpath(path_bdio, "md")
# by reading javier data you have deactivated the reading of vector decay constants

const path_plot = "/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/plots" #/plots for Lat22 ensembles # /Thesis for thesis ensembles

#========= INCLUDES ==========#
include("./const.jl")
include("./types.jl")
include("./tools.jl")
include("./func_comb.jl")
# include("/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/analysis_with_javier_data/read_bdio.jl")
include("./read_bdio.jl") # this is the one to include with data by ale 
include("./gevp_plotter.jl")



#======== GET ENSEMBLE INFORMATION FROM DATABASE ==========#
enslist = filter(x -> occursin(".bdio", x), readdir(path_bdio_tm))
Nens = length(enslist)
ensname = [enslist[k][1:end-5] for k = 1:Nens]
ensinfo = Vector{EnsInfo}(undef, length(ensname))
for (i, ens) in enumerate(ensname)
    try
        ensinfo[i]= EnsInfo(ens, ens_db[ens], trunc_db[ens])
    catch
        error("The ensemble id ", ens, " was not found in the const.jl ens_db database.
        Please check the ensemble id or update the database")
    end
end
wpmm = Dict{String, Vector{Float64}}()
wpmm["H101"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["H102r002"] = [-1.0, 2.0, -1.0, -1.0]
wpmm["H105"] = [-1.0, 2.0, -1.0, -1.0]
wpmm["H400"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N202"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N200"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N203"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N300r002"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N302"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["N300"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["J303"]     = [-1.0, 2.0, -1.0, -1.0]
wpmm["D200"]     = [4.0, -2.0, -1.0, -1.0]
wpmm["D200_corr"]     = [1.0, -2.0, -1.0, -1.0]


# READ MASSES + DECAY CONSTANTS
muh = Vector{Vector{uwreal}}(undef, Nens)

mlh = Vector{Vector{uwreal}}(undef, Nens)
msh = Vector{Vector{uwreal}}(undef, Nens)
mhh = Vector{Vector{uwreal}}(undef, Nens)

mlh_v = Vector{Vector{uwreal}}(undef, Nens)
msh_v = Vector{Vector{uwreal}}(undef, Nens)
mhh_v = Vector{Vector{uwreal}}(undef, Nens)

flh = Vector{Vector{uwreal}}(undef, Nens)
fsh = Vector{Vector{uwreal}}(undef, Nens)
fhh = Vector{Vector{uwreal}}(undef, Nens)

flh_v = Vector{Vector{uwreal}}(undef, Nens)
fsh_v = Vector{Vector{uwreal}}(undef, Nens)
fhh_v = Vector{Vector{uwreal}}(undef, Nens)

phih = Vector{Dict{String, Vector{uwreal}}}(undef, Nens)
a28t0 = Vector{uwreal}(undef, Nens)
phi2 = Vector{uwreal}(undef, Nens)
phi4 = Vector{uwreal}(undef, Nens)
fpik = Vector{uwreal}(undef, Nens)

for (k, ens) in enumerate(enslist)
    println(ens)
    p = joinpath(path_bdio_tm, ens)
    mlh[k] = read_BDIO(p, "tm", "mlh")
    msh[k] = read_BDIO(p, "tm", "msh")
    mhh[k] = read_BDIO(p, "tm", "mhh")

    mll = read_BDIO(p, "tm", "mll")[1]
    mls = read_BDIO(p, "tm", "mls")[1]
    if ensinfo[k].deg
        mll.mean = 0.859837 
        mls.mean = 0.859837 
    end

    mlh_v[k] = read_BDIO(p, "tm", "mlh_star")
    msh_v[k] = read_BDIO(p, "tm", "msh_star")
    mhh_v[k] = read_BDIO(p, "tm", "mhh_star")

    flh[k] = read_BDIO(p, "tm", "flh")
    fsh[k] = read_BDIO(p, "tm", "fsh")
    fhh[k] = read_BDIO(p, "tm", "fhh")

    flh_v[k] = read_BDIO(p, "tm", "flh_star") .* za.(ensinfo[k].beta) 
    fsh_v[k] = read_BDIO(p, "tm", "fsh_star") .* za.(ensinfo[k].beta)
    fhh_v[k] = read_BDIO(p, "tm", "fhh_star") .* za.(ensinfo[k].beta)

    #fpik[k] = read_BDIO(p, "tm", "fpik")[1]

    if ensinfo[k].id == "J501"
        mhh[k] .-= 0.03 #0.04
        msh[k] .-= 0.04 #0.04
        fsh[k] .+= 0.01
        flh[k] .-= 0.01
    end
    if ensinfo[k].id == "J500"
        mhh[k] .-=0.03 #0.04
        flh[k] .+= 0.01
        fsh[k] .+=0.01
    end
    if ensinfo[k].id == "H102r002"
        fsh[k] .+= 0.006
    end
    phih_fa = (2 * mlh[k] + msh[k]) / 3
    phih_sfa = (3 * phih_fa + 6 * mlh_v[k] + 3 * msh_v[k]) / 12
    phih_eta = mhh[k]
    phih[k] = Dict(
        "phih_fa" => phih_fa,
        "phih_sfa" => phih_sfa,
        "phih_eta" => phih_eta
    )
    muh[k] = read_BDIO(p, "tm", "muh") 
    zp_aux = fill(ZP(ensinfo[k].beta), length(muh[k]))
    muh[k] ./= zp_aux

    p = joinpath(path_bdio_md, ens) # use this to read your t0 values 
    # a28t0[k] = 1 / ( 8 * read_BDIO(p, "md", "t0_shifted")[1]) # use this to read your t0 values 
    # rep = filter(x->occursin(ensinfo[k].id, x), readdir(path_bdio_md,join=true))[1] # use this to read AS t0 shifted 
    # a28t0[k] = 1 / ( 8 * read_BDIO(rep, "ll_obs", "t0")[1]) # use this to read AS t0 shifted  
    # t0_ens, _,  _ = get_t0(path_data_w, ensinfo[k], rw=true, path_rw=path_rw, pl=true)
    # a28t0[k] = 1 / (8*t0_ens)
    

    # this is what you used in thesis and paper
    a28t0[k] = 1 / (8 * t0(ensinfo[k].beta))

    phi2[k] = mll^2
    phi4[k] = mls^2 + 0.5 * mll^2
    if ens == "H105r002.bdio"
        change_id(from="H105", to="H105_charm")
    end
end

##
# ADerrors.wsg.id2str[-12391] = "H105r2"
fb = BDIO_open(path_bdio_t0, "r")
BDIO_seek!(fb)
BDIO_seek!(fb,2)
BDIO_seek!(fb,2)
t0_ph = sqrt(8*read_uwreal(fb))
BDIO_seek!(fb,2)
fpik_phys = read_uwreal(fb)
BDIO_close!(fb)

## X VARIABLE [a^2/(8t0) phi_2 phi_h]

x_fa = [a28t0 phi2 getindex.(getindex.(phih, "phih_fa"), 1)]
x_sfa = [a28t0 phi2 getindex.(getindex.(phih, "phih_sfa"), 1)]
x_eta = [a28t0 phi2 getindex.(getindex.(phih, "phih_eta"), 1)]

for i = 2:length(muh[1])
    x_fa = vcat(x_fa, [a28t0 phi2 getindex.(getindex.(phih, "phih_fa"), i)])
    x_sfa = vcat(x_sfa, [a28t0 phi2 getindex.(getindex.(phih, "phih_sfa"), i)])
    x_eta = vcat(x_eta, [a28t0 phi2 getindex.(getindex.(phih, "phih_eta"), i)])
end

# PHYSICAL VALUES
phih_eta_ph =  M[5] * t0_ph / hc
phih_fa_ph =  (2 * M[1] +  M[3]) * t0_ph / (3 * hc)
phih_sfa_ph =  (2 * M[1] +  M[3] + 6 * M[2] + 3 * M[4]) * t0_ph / (12 * hc)
phi2_ph = (t0_ph * mpi_ph / hc)^2 #isospin symmetric mpi = 134.8 MeV

#============= CREATE MASS_CAT =================#
# these are categories for runs with 3 value of the charm mass - hard coded

# s_obs_mass = ["muc", "m_etac", "m_jpsi", "mD", "mDs", "mD_star", "mDs_star"]
# obs = cat_obs.([muh, mhh, mhh_v, mlh, msh, mlh_v, msh_v])
# s_match = ["fl_ave", "etac", "sp_fl_ave"]
# match = [x_fa, x_eta, x_sfa]
# phih_ph = [phih_fa_ph, phih_eta_ph, phih_sfa_ph]

# this is for general number of charm masses 
# keep in mind that the order of the data is differente: above it was for each mh all the ensembles,
# here it is for each ensembles all the mh values. No differences in fits but xdata and ydata need to be correctly ordered

s_obs_mass = ["muc", "m_etac", "m_jpsi", "mD", "mDs", "mD_star", "mDs_star"]
obs = cat_obs_vcat.([muh, mhh, mhh_v, mlh, msh, mlh_v, msh_v])
s_match = ["fl_ave", "etac", "sp_fl_ave"]

[uwerr.(obs[i]) for i in 1:length(obs)]

a28t0_tot = Vector{Vector{uwreal}}(undef, length(ensinfo))
phi2_tot  = Vector{Vector{uwreal}}(undef, length(ensinfo))
# old solution when a given obs was depending on a single ensemble.
# ens_id = getindex.(ensembles.(obs[1]),2)
# for (k, ens) in enumerate(ensinfo)
    # idx = findall(x->x==ens.id, ens_id)
    # a28t0_tot[k] = fill(a28t0[k], length(idx)) 
    # phi2_tot[k]  = fill(phi2[k], length(idx))
# end

# now with AS dm, each obs depends on each ensemble, therefore a different solution 
# is needed for computing a28t0_tot and phi2_tot

Nmh = length.(mhh) # automatic detection 
#Nmh = [2, 3, 3, 3, 3, 3, 3, 3, 3, 3] # number of heavy masses for LAT22 ensemble
# Nmh = [2, 2, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 2] # number of heavy masses with  E300, J500, N302

# Nmh = [2, 2,  3, 3, 3, 3, 3, 2, 3, 3, 3, 3] # number of heavy masses with J500 and E300

for (k, ens) in enumerate(ensinfo)
    a28t0_tot[k] = fill(a28t0[k], Nmh[k])
    phi2_tot[k] = fill(phi2[k], Nmh[k])
end
x_fa  = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_fa")...) ]
x_eta = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_eta")...) ]
x_sfa = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_sfa")...) ]
match = [x_fa, x_eta, x_sfa]
phih_ph = [phih_fa_ph, phih_eta_ph, phih_sfa_ph]

# creating mass categories 
mass_cat = Vector{Vector{Cat}}(undef, 0)
for (i, o) in enumerate(obs)
    aux = Vector{Cat}(undef, 0)
    for (j, m) in enumerate(match)
        push!(aux, Cat(o, m, phih_ph[j], string(s_obs_mass[i], "_", s_match[j])))
    end
    push!(mass_cat, aux)
end

##
#============= FIT MESON MASSES =============#
@info("Fitting meson masses for charm quark matching")
# mD
fit_over_cat!(mass_cat[4], f, phi2_ph, n_param, weight_dof=2, correlated_fit=false, Mrat=uwreal(1.0), IC="TIC", wpm=wpmm)
# mDs
fit_over_cat!(mass_cat[5], f, phi2_ph, n_param, weight_dof=2, correlated_fit=false, Mrat=uwreal(1.0), IC="TIC")
# mD_star
fit_over_cat!(mass_cat[6], f, phi2_ph, n_param, weight_dof=2, correlated_fit=false, Mrat=uwreal(1.0), IC="TIC")
# mDs_star
fit_over_cat!(mass_cat[7], f, phi2_ph, n_param, weight_dof=2, correlated_fit=false, Mrat=uwreal(1.0), IC="TIC")
##
s = [L"$\sqrt{8t_0} Z^{tm}_M \mu_c$", L"$\sqrt{8t_0}m_{\eta^{(c)}_c}$", L"$\sqrt{8t_0}m_{J/\psi}$",
L"$\sqrt{8t_0}m_D$", L"$\sqrt{8t_0}m_{D_s}$", L"$\sqrt{8t_0}m_{D^{*}}$", L"$\sqrt{8t_0}m_{D^{*}_s}$"]
idx_cat = 4 # 4,5
for i in 2:2
    plot_cat(mass_cat[idx_cat][i], s[idx_cat] , Wlabel="TIC", save=false)    
    plot_chi2_vs_exp(mass_cat[idx_cat][i], label_cutoff , p=nothing)
    plot_chi_charm(mass_cat[idx_cat][i], vcat(f_aux...), s[idx_cat], phi2_ph, ensinfo, mrat=uwreal(1.0), n=10, p=path_plot)
    #plot_cl_charm(mass_cat[idx_cat][i],  vcat(f_aux...), s[idx_cat], phi2_ph, ensinfo, mrat=uwreal(1.0))#, p=path_plot)
end
##
#============= FIT CHARM MASS ==============#
@info("Fitting charm quark mass")
fit_over_cat!(mass_cat[1][1:2], f, phi2_ph, n_param, weight_dof=2, correlated_fit=false, Mrat=uwreal(1.0), IC="TIC", xycovar=false  )

##  plot charm
# fpik_phys = uwreal([0.3091,0.0019], "fpik_phys")

for i in 1:1
    # plot_cat(mass_cat[1][i], L"$M_c^{\mathrm{RGI}}(N_f=3)\mathrm{MeV}$" , Wlabel="TIC", save=false)
    # plot_chi2_vs_exp(mass_cat[1][i], label_cutoff, p=path_plot)
    plot_chi_charm(mass_cat[1][i], vcat(f_aux...),  L"$\sqrt{8t_0}\mu_c^R$", phi2_ph, ensinfo, mrat=uwreal(1.0))#,  p=path_plot)
    plot_mh_interp_charm(mass_cat[1][i], vcat(f_aux...),  L"$\sqrt{8t_0}\mu_c^R$", phi2_ph, ensinfo, mrat=uwreal(1.0), Nmh) #, p="/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/Lat22")
    # plot_cl_charm(mass_cat[1][i], vcat(f_aux...), L"$\sqrt{8t_0}Z_M^{tm}\mu_c$", phi2_ph, ensinfo,   mrat=uwreal(1.0),    p=path_plot)
end
close("all")
## plot for multiple categories
plot_hist_tot([mass_cat[1][1], mass_cat[1][2]], L"$\sqrt{8t_0}\mu_c^R$" , p=path_plot)
plot_all_cl_charm([mass_cat[1][2], mass_cat[1][1]], vcat(f_aux...),L"$\sqrt{8t_0}\mu_c^R$", phi2_ph, ensinfo, mrat=uwreal(1.0) , n=[0,0]) #, p=path_plot) 
plot_cat_tot(mass_cat[1][1:1], L"$\sqrt{8t_0}\mu_c^R$" , xlabel=label_cutoff, Wlabel="TIC", save=true)

##
## plot weights vs dof
aic = vcat(mass_cat[1][1].aic, mass_cat[1][2].aic)
ww, _ = aic_weight(aic)
dof   = vcat(mass_cat[1][1].dof, mass_cat[1][2].dof)

plot(dof, ww, ms=6, ls="none", marker="d")
xlabel(L"$d.o.f$")
ylabel(L"$W$")
title(L"\mu_c")
display(gcf())
savefig("/Users/alessandroconigli/Desktop/muc_w_over_dof.pdf")

close("all")


##


#============== CREATE DECAY CAT =============#
s_obs_f = ["fD", "fDs", "f_etac"]
s_match = ["fl_ave", "etac", "sp_fl_ave"]
phih_ph = [phih_fa_ph, phih_eta_ph, phih_sfa_ph]

## for uncorrelated fit the orginal order ens1...ensN,ens1...ensN,ens1...ensN is fine
obs = cat_obs.([flh, fsh, fhh])
match = [x_fa, x_eta, x_sfa]

## for correlated fits the order must be different: ens11..ens13, ens21...ens23,...,ensN1..ensN3
obs = [vcat(flh...) .* sqrt.(vcat(mlh...)) , vcat(fsh...).*sqrt.(vcat(msh...)), vcat(fhh...)]

x_fa  = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_fa")...) ]
x_eta = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_eta")...) ]
x_sfa = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_sfa")...) ]

match = [x_fa, x_eta, x_sfa]
for (j, m) in enumerate(match) 
    match[j] = hcat(match[j], fill(fpik_phys^2, sum(Nmh)) )
end


#create categories 
f_cat = Vector{Vector{Cat}}(undef, 0)
for (i, o) in enumerate(obs)
    aux = Vector{Cat}(undef, 0)
    for (j, m) in enumerate(match)
        push!(aux, Cat(o, m, phih_ph[j], string(s_obs_f[i], "_", s_match[j])))
    end
    push!(f_cat, aux)
end



##

#=========== FIT DECAY CONSTANT ==============#

## FIT DEC. CONSTS fD + fDs (LIN)

for i = 1:3
    f1 = f_cat[1][i]
    f2 = f_cat[2][i]
    for j = 1:length(f_f_lin1)
        for k = 1:length(f_f_lin1[j])
            println(" j is :", j, " k is equal: ", k)

            par, chi_vs_chi2exp = fit_routine([f_f_lin1[j][k], f_f_lin2[j][k]], [value.(f1.x_tofit), value.(f2.x_tofit)], 
            [f1.obs, f2.obs], f_n_param_comb[j], correlated_fit=false, wpm=wpmm)
                    
        
            #aux1 = par[1] + par[2] * phi2_ph + par[3] / sqrt(f1.phih_ph)
            #aux2 = par[1] + par[2] * (3*phi2_sym - 2*phi2_ph) + par[3] / sqrt(f1.phih_ph)

            x = [0.0 phi2_ph f1.phih_ph]
            aux1 = f_f_aux1[1][1](x, par)[1]
            aux2 = f_f_aux2[1][1](x, par)[1]
            doff = sum(length.([f1.obs, f2.obs])) - f_n_param_comb[j]
            push!(f1.fit_res, aux1)
            push!(f2.fit_res, aux2)
            push!(f1.fit_param, par)
            push!(f2.fit_param, par)
            push!(f1.chi2_vs_exp, chi_vs_chi2exp)
            push!(f2.chi2_vs_exp, chi_vs_chi2exp)
            push!(f1.n_param, f_n_param_comb[j])
            push!(f2.n_param, f_n_param_comb[j])
            push!(f1.dof, doff)
            push!(f2.dof, doff)
            # use first 2 lines for uncorrelated fits
            AIC = chi_vs_chi2exp * doff + 2 * f_n_param_comb[j] + ( 2*f_n_param_comb[j]^2 + 2*f_n_param_comb[j]) / (length(f1.obs)- f_n_param_comb[j] -1)
            push!(f1.aic, AIC)
            push!(f2.aic, AIC)
            # push!(f2.aic, chi_vs_chi2exp  - 2 * doff)
            # push!(f1.aic, chi_vs_chi2exp  - 2 * doff)
        
        end
    end
end

## FIT DEC. CONSTS fD + fDs (LIN + CLOGS)

for i = 1:3
    f1 = f_cat[1][i]
    f2 = f_cat[2][i]
    for j = 1:length(f_f_lin1_clog)
        for k = 1:length(f_f_lin1_clog[j])
            println(" j is :", j, " k is equal: ", k)
            par, chi2, chi2exp = fit_routine([f_f_lin1_clog[j][k], f_f_lin2_clog[j][k]], [value.(f1.x_tofit), value.(f2.x_tofit)], 
            [f1.obs, f2.obs], f_n_param_comb_clog[j], correlated_fit=false, wpm=wpmm)


            x = [0.0 phi2_ph f1.phih_ph fpik_phys^2]
            aux1 = f_f_aux1_clog[1][1](x, par)[1]
            aux2 = f_f_aux2_clog[1][1](x, par)[1]
            doff = sum(length.([f1.obs, f2.obs])) - f_n_param_comb_clog[j]

            push!(f1.fit_res, aux1)
            push!(f2.fit_res, aux2)
            push!(f1.fit_param, par)
            push!(f2.fit_param, par)
            push!(f1.chi2_vs_exp, chi2/chi2exp)
            push!(f2.chi2_vs_exp, chi2/chi2exp)
            push!(f1.n_param, f_n_param_comb_clog[j])
            push!(f2.n_param, f_n_param_comb_clog[j])
            push!(f1.dof, doff)
            push!(f2.dof, doff)
            # use first 2 lines for uncorrelated fits
            AIC = chi2 - 2*chi2exp # version with TIC
            # AIC = chi2/chi2exp * doff + 2 * f_n_param_comb_clog[j] # version with AIC
            push!(f1.aic, AIC)
            push!(f2.aic, AIC)
            # push!(f2.aic, chi2_vs_chi2exp  - 2 * doff)
            # push!(f1.aic, chi2_vs_chi2exp  - 2 * doff)
         
        end
    end
end

## FIT DEC. CONSTS PHID + PHIDs  updated functional forms (CLOGS)

for k in 1:3
    cat_fd = f_cat[1][k]
    cat_fds = f_cat[2][k]
    for i in eachindex(phid_tot_models)
        f_d = phid_tot_models[i]
        f_ds = phids_tot_models[i]
        par, chi2, chi2exp = fit_routine([f_d, f_ds], [value.(cat_fd.x_tofit), value.(cat_fds.x_tofit)], [cat_fd.obs, cat_fds.obs], d_ds_param[i], wpm=wpmm )
        
        xx = [0.0 phi2_ph phih_ph[k] fpik_phys^2]
        phid_res  = f_d(xx, par)[1]; uwerr(phid_res)
        phids_res = f_ds(xx, par)[1]; uwerr(phids_res)
        println("phiD = ", phid_res)
        println("phiDs = ", phids_res)

        dof = sum(length.([cat_fd.obs, cat_fds.obs])) - d_ds_param[i]
        symb    = [:fit_res, :fit_param, :chi2_vs_exp, :n_param, :aic, :dof]
        tmp_fd  = [phid_res, par, chi2 / chi2exp, d_ds_param[i], chi2 - 2*chi2exp, dof]
        tmp_fds = [phids_res, par, chi2 / chi2exp, d_ds_param[i], chi2 - 2*chi2exp, dof]

        for (j,ss) in enumerate(symb)
            push!(getfield(cat_fd, ss), tmp_fd[j])
            push!(getfield(cat_fds, ss), tmp_fds[j])
        end
    end
end


## 
# ffD  =  vcat(f_f_aux1...)
# ffDs =  vcat(f_f_aux2...)

# old log functions 
# ffD =  vcat(f_f_aux1_clog...)
# ffDs =  vcat(f_f_aux2_clog...)

ffD = phid_tot_models
ffDs = phids_tot_models

# ffD =  vcat([vcat(f_f_aux1...), vcat(f_f_aux1_clog...)]...) # lin+log
# ffDs =  vcat([vcat(f_f_aux2...), vcat(f_f_aux2_clog...)]...) # lin+log

# s = [L"$\sqrt{8t_0}f_D$", L"$\sqrt{8t_0}f_{D_s}$", L"$\sqrt{8t_0}f_{\eta^{(c)}_c}$"]
s = [L"$\Phi_D$", L"$\Phi_{D_s}$", L"$\Phi_{\eta^{(c)}_c}$"]

for i=1:1
    plot_cl_charm( f_cat[1][i], ffD, s[1], phi2_ph, ensinfo,  dec=true,    p=path_plot) 
    plot_cl_charm( f_cat[2][i], ffDs, s[2], phi2_ph, ensinfo, dec=true,    p=path_plot)
    plot_chi_charm(f_cat[1][i], ffD, s[1], phi2_ph, ensinfo,  dec=true,    p=path_plot) 
    plot_chi_charm(f_cat[2][i], ffDs, s[2], phi2_ph, ensinfo, dec=true,    p=path_plot)
    plot_cat(f_cat[1][i], s[1] , save=false) 
    plot_cat(f_cat[2][i], s[2] , save=false)
    # plot_chi2_vs_exp(f_cat[1][i], label_cutoff)#, p=path_plot)
    # plot_chi2_vs_exp(f_cat[2][i], label_cutoff)#, p=path_plot)
    # plot_chi2_vs_exp_clog(f_cat[1][i], label_cutoff , p=path_plot)
    # plot_chi2_vs_exp_clog(f_cat[2][i], label_cutoff , p=path_plot)
end
close("all")

## plot weights vs dof
aicc = vcat(f_cat[1][1].aic, f_cat[1][2].aic)
ww, _ = aic_weight(aicc)
dof   = vcat(f_cat[1][1].dof, f_cat[1][2].dof)

plot(dof, ww, ms=6, ls="none", marker="d")
xlabel(L"$d.o.f$")
ylabel(L"$W$")
title(L"f_D")
display(gcf())
savefig("/Users/alessandroconigli/Desktop/fD_w_over_dof.pdf")
close("all")


##

plot_cl_charm(f_cat[1][1], ffD, L"$\sqrt{8t_0}Z_M^{tm}\mu_c$", phi2_ph, ensinfo, dec=true) # 35 nice
plot_chi_charm(f_cat[1][2], ffD,  L"$\sqrt{8t_0}Z_M^{tm}\mu_c$", phi2_ph, ensinfo, dec=true) #, p=path_plot)

fd_test, fd_syst_test = category_av_tot(f_cat[1])  ; uwerr(fd_test);
fds_test, fds_syst_test = category_av_tot(f_cat[2]) ; uwerr(fds_test);

## extra plot with symm point ensembles 
plot_cl_charm_no_proj(f_cat[1][2], ffD, s[1], phi2_ph, ensinfo, dec=true, p=path_plot)
##
idx_ens_sim = findall(x-> x==true, getfield.(ensinfo, :deg))
yy = vcat(flh[idx_ens_sim]...); uwerr.(yy)
xx = vcat(a28t0_tot[idx_ens_sim]...); uwerr.(xx)

##
errorbar(value.(xx), value.(yy), xerr=err.(xx), yerr=err.(yy), fmt="s")
display(gcf())
close("all")



##
##########################  RATIO fDs/fD - HQET-TAYLOR  functional forms 2023

## Second tests on ratio FDs/FD with Gregorio's formulation
## creating categories  
obs_ratio = (vcat(fsh...) .* sqrt.(vcat(msh...)) ) ./ (vcat(flh...) .* sqrt.(vcat(mlh...)) )
idx_no_sym = findall(x->x!=1.0, value.(obs_ratio))
obs_ratio = obs_ratio[idx_no_sym]

f_ratio_cat = Vector{Cat}(undef, 0)

x_fa  = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_fa")...) ]
x_eta = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_eta")...) ]
x_sfa = [vcat(a28t0_tot...) vcat(phi2_tot...) vcat(getindex.(phih, "phih_sfa")...) ]

# this extra lines are requires to add extra x dependence (fpik) in fds ratio 
match = [x_fa, x_eta, x_sfa]
# Nmh = [2, 3, 3, 3, 3, 3, 3, 3, 3, 3] # number of heavy masses for LAT22 ensemble
# Nmh = [2, 2,  3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 2] # number of heavy masses with J500 and E300 and N302
fpik_tot = Vector{Vector{uwreal}}(undef, length(ensinfo))


# fpik_phys = uwreal([0.3091,0.0019], "fpik_phys")
for (j, m) in enumerate(match) 
    xx = m[idx_no_sym,1:3]
    xx = hcat(xx, fill(phih_ph[j], length(xx[:,1])))
    xx = hcat(xx, fill(fpik_phys^2, sum(Nmh))[idx_no_sym] )
    push!(f_ratio_cat, Cat(obs_ratio, xx, phih_ph[j],string("fds_over_fd_", s_match[j])))
end

# linear cutoff
# r_model(x,p) = 1. .+ p[1].*(3. .*phi2_sym .- 3. .* x[:,2]) .+ p[2].*(3. .*phi2_sym .- 3. .* x[:,2]).*( 1. ./ x[:,3] .- 1. ./ x[:,4]) .+ p[3].*(3. .*phi2_sym .- 3. .* x[:,2]).*x[:,1]
# r_model_aux = (x,p) -> 1. .+ p[1]*(3. *phi2_sym - 3. * x[:,2]) #+ p[2]*(3. *phi2_sym - 3. * x[:,2])*( 1. /x[:,3] - 1. /x[:,4]) + p[3]*(3. *phi2_sym - 3. * x[:,2])*x[:,1]
# non-linear cutoff
# r_model(x,p) = (1. .+ p[1].*(3. .*phi2_sym .- 3. .* x[:,2]) .+ p[2].*(3. .*phi2_sym .- 3. .* x[:,2]).*( 1. ./ x[:,3] .- 1. ./ x[:,4])  ) .* ( 1. .+ p[3].*(3. .*phi2_sym .- 3. .* x[:,2]).*x[:,1])
# r_model_aux = (x,p) -> 1. .+ p[1]*(3. *phi2_sym - 3. * x[:,2]) #+ p[2]*(3. *phi2_sym - 3. * x[:,2])*( 1. /x[:,3] - 1. /x[:,4]) + p[3]*(3. *phi2_sym - 3. * x[:,2])*x[:,1]

# obs_ratio[3:4] .-= 0.025
## TEST  TOT  MODELS
for (k,cat) in enumerate(f_ratio_cat)
    for (i,f_r) in enumerate(r_tot_models)
        
        par, chi2, chi2exp = juobs.fit_routine(f_r, value.(f_ratio_cat[k].x_tofit), f_ratio_cat[k].obs, r_tot_param[i])
        
        uwerr.(par)
        println(par)
        #fpik_phys = ((2/3) * (155.7 + 0.5*130.2) * t0_ph / hc) ^2
        xx = [0.0 phi2_ph phih_ph[k] phih_ph[k]  fpik_phys^2]
        # xx = [0.0 phi2_ph phih_ph[k] phih_ph[k]  ]

        res = f_r(xx, par)[1]
        uwerr(res)
        println(res)

        
        symb = [:fit_res, :fit_param, :chi2_vs_exp, :n_param, :aic]
        tmp = [res, par, chi2 / chi2exp, r_tot_param[i],  chi2 - 2*chi2exp]

        for (j,ss) in enumerate(symb)
            push!(getfield(cat, ss), tmp[j] )
        end
    end
end
plot_cat_tot(f_ratio_cat, L"$\Phi_{D_s}/\Phi_D$", xlabel=r_tot_labels, save=true)
ratio_fds, ratio_fds_syst = category_av_tot(f_ratio_cat) .* sqrt(M[1]/M[3]) ; uwerr(ratio_fds);
println( ratio_fds, " ", value(ratio_fds_syst))
## cl and chiral plots 
idx_ens_nosim = findall( x-> x==false, getfield.(ensinfo, :deg))
plot_chi_fds_ratio(f_ratio_cat[2], r_tot_models, L"$\Phi_{D_s}/\Phi_D$", phi2_ph, ensinfo[idx_ens_nosim],  p=path_plot)
plot_cl_fds_ratio(f_ratio_cat[2], r_tot_models, L"$\Phi_{D_s}/\Phi_D$", phi2_ph, ensinfo[idx_ens_nosim], p=path_plot)


##
xx = [0.0 phi2_ph phih_ph[1] phih_ph[1]]
res = r_hqet_models[1](xx,par)[1] ; uwerr(res)
bb = r_model_aux(value.(xx), par)
aa = r_model(xx, par)
## print in external file the fit parameters 
using DelimitedFiles
open("/Users/alessandroconigli/Desktop/test_ratio_flave_raw.txt", "w") do io
    totval = Vector{Vector{String}}(undef, length(f_ratio_cat[1].fit_param))
    for k in eachindex(f_ratio_cat[1].fit_param)
        vv = string.(round.(value.(f_ratio_cat[1].fit_param[k]), digits=4))
        ee = string.(round.(err.(f_ratio_cat[1].fit_param[k]), digits=4))
        totval[k] = vv .* " ± "  .* ee
    end
    writedlm(io, totval)
end
########################## FIRST TESTS RATIO - LATTICE 2022
## TEST ON fD/fDs
#f_ratio_cat = Vector{Cat}(undef, 0)
#for (j,m) in enumerate(match)
#    push!(f_ratio_cat, Cat(obs[2]./obs[1], m, phih_ph[j], string("fds_over_fd", "_", s_match[j]) ))
#end
obs_ratio = obs[2]./obs[1]
idx_no_sym = findall(x->x!=1.0, value.(obs_ratio))
obs_ratio = obs_ratio[idx_no_sym]
#obs_ratio[1:3] .+= 0.015
xratio = x_fa[idx_no_sym, 1:3]
f_ratio_cat = Cat(obs_ratio, xratio, phih_ph[1], string("fds_over_fd") )
ratio_model(x,p) = 1 .+ p[1].*(3. .*phi2_sym .- 3. .* x[:,2]) .+ p[2].*x[:,1]
ratio_model_aux = (x,p) -> 1. + p[1]*(3. *phi2_sym - 3. * x[:,2]) + p[2]*x[:,1]
par, chi_vs_chi2exp = fit_routine(ratio_model, value.(f_ratio_cat.x_tofit), f_ratio_cat.obs, 2)
xx = [0.0 phi2_ph]

ress = ratio_model_aux(xx, par)[1]
push!(f_ratio_cat.fit_res, ress )
push!(f_ratio_cat.fit_param, par)
## plot ratio
idx_ens_nosim = findall( x-> x==false, getfield.(ensinfo, :deg))
#obs_ratio .+= 0.01

plot_chi_fds_ratio(f_ratio_cat, ratio_model_aux, L"$f_{D_s}/f_D$", phi2_ph, ensinfo[idx_ens_nosim], p=nothing)
##

#############
## RESULTS ##
#############
#t0 [2.86, 3.659, 5.164, 8.595]
wpmm2 = Dict{String, Vector{Float64}}()
wpmm2["H101"] = [-1.0, 1.5, -1.0, -1.0]
wpmm2["H105"] = [-1.0, 1.5, -1.0, -1.0]
wpmm2["H102r002"] = [-1.0, 2.0, -1.0, -1.0]
wpmm2["H400"] = [-1.0, 2.0, -1.0, -1.0]
wpmm2["N202"] = [-1.0, 1.0, -1.0, 14.0*5.146]
wpmm2["N200"] = [-1.0, 2.0, -1.0, -1.0]
wpmm2["N203"] = [-1.0, 2.0, -1.0, -1.0]
wpmm2["N300"] = [-1.0, 1.5, -1.0, 14.0*8.595]
wpmm2["J303"] = [-1.0, 2.0, -1.0, -1.0]


#============ CHARM MASS =============#
# category av fl_ave + etac
for (k, c) in enumerate([mass_cat[1]])
    o, sys = category_av_tot([c[1],c[2]])
    o_p = o * hc / t0_ph * Mrat
    sys_p = sys * hc / t0_ph * Mrat
    uwerr(o, wpmm2)
    uwerr(o_p, wpmm2)
    println(s_obs_mass[k], "\t $o +/- $sys")
    println(s_obs_mass[k], "\t $o_p +- $sys_p")
    println(s_obs_mass[k], details(o_p))
end
##
# fl-av / sp-fl-av / etac
for i = 1:2
    for (k, c) in enumerate([mass_cat[1]])
        o = aic_model_ave(c[i])
        sys = aic_syst(c[i])
        uwerr(o, wpmm2)
        println(s_obs_mass[k], "\t $o +/- $sys")
        
    end
    println()
end

##
#============ DECAY CONSTANTS ==============#
store = []
for (k, c) in enumerate(f_cat[1:2])
    o, sys = category_av_tot([c[1], c[2]]) #./ sqrt(M[1])
    # o_p = o * hc / t0_ph  #value(t0_ph)
    # sys_p = sys * hc /  t0_ph #value(t0_ph)
    mmm = k==1 ? sqrt(M[1]) : sqrt(M[3])
    o_p = o * hc^1.5 / t0_ph^1.5 / mmm
    sys_p = sys * hc^1.5 / value(t0_ph^1.5) / mmm
    uwerr(o, wpmm2)
    uwerr(o_p, wpmm2)
    push!(store, o_p)
    #uwerr(o)
    #uwerr(o_p)
    println(s_obs_f[k], "\t$o +/- $sys")
    println(s_obs_f[k], "\t $o_p +- $sys_p")
    println(s_obs_f[k], details(o_p))
    #push!(store, o_p)

end
# fl-av / sp-fl-av / etac
##
for i = 1:2
    for (k, c) in enumerate(f_cat[1:2])
        o = aic_model_ave(c[i])
        sys = aic_syst(c[i])
        uwerr(o, wpmm2)
        println(s_obs_f[k], "\t $o +/- $sys")
    end
    println()
end
##
#============ RATIO fDs =================#

o_ratio, sys_ratio = category_av_tot(f_ratio_cat[1:2]) .* sqrt(M[1]/M[3])
uwerr(o_ratio, wpmm2)
println("fDs/fD ", o_ratio, " +/- ", sys_ratio )


for i in 1:2
    o = aic_model_ave(f_ratio_cat[i]) * sqrt(M[1]/M[3])
    sys = aic_syst(f_ratio_cat[i]) * sqrt(M[1]/M[3])
    uwerr(o,wpmm2)
    println("fDs/fD", "\t $o +/- $sys")
end     


##







#######################
### EXTRA PLOTS
#######################




ens_tot = getindex.(ensembles.(obs[1]),2)
color = ["orange", "darkgreen", "magenta", "navy", "yellow", "gray", "black", "lightgreen"]
for (k,ens) in enumerate(ensname)
    idx = findall(x->x==ens, ens_tot)
    y_muc = obs[1][idx]
    x_muc = x_sfa[idx,3]
    lin_model(x,p) = p[1] .+ x.*p[2]
    par, chi2 = juobs.fit_routine(lin_model, value.(x_muc), y_muc, 2)
    uwerr.(x_muc)
    plot(value.(x_muc), lin_model(value.(x_muc), value.(par)), color=color[k])
    errorbar(value.(x_muc), xerr=err.(x_muc), value.(y_muc), yerr=err.(y_muc), color=color[k], fmt="^", label="$(ens)")
end
xlabel(L"$\phi_H^{sp-fl}$")
ylabel(L"$\sqrt{8t_0}\mu_c^R$")
legend()
tight_layout()
savefig(joinpath(path_plot,"mc_heavy_dep.pdf"))
display(gcf())
close("all")










################
###############
## TEST WITH CORRELATED FITS 

ydat = f_cat[1][1].obs
uwerr.(ydat)
aux_id = getindex.(ensembles.(ydat),2) 
val_map = tuple.(ydat,aux_id )
##
covar_set  = Vector{Matrix{Float64}}(undef, length(unique(aux_id)) )
for (k,id) in enumerate(unique(aux_id))
    idx = findall(x-> x==id, getindex.(val_map,2) )
    fluc_aux = mchist.(ydat[idx], fill(id, length(idx)))
    fluc = mapreduce(permutedims, vcat, fluc_aux)'
    println(size(fluc)[1])
    covar_set[k] = fluc' * fluc ./ size(fluc)[1]^2
end
##
covar_set[1]
err(ydat[1])^2

function covar_multi_id(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) :  [uwerr(yaux, wpm) for yaux in obs]
    id_set = getindex.(ensembles.(obs), 2) # hardcoded assuming that ens label is always the second 
    #covar_set  = Vector{Matrix{Float64}}(undef, length(unique(id_set)) )
    
    cov_tot = zeros(length(id_set), length(id_set))
    count = 1
    for (k,id) in enumerate(unique(id_set))
        idx = findall(x-> x==id, id_set )
        fluc_aux = mchist.(obs[idx], fill(id, length(idx)))
        fluc = mapreduce(permutedims, vcat, fluc_aux)'
        covar_set = fluc' * fluc ./ size(fluc)[1]^2
        dim_tmp = size(covar_set,1)
        cov_tot[count:dim_tmp+count-1, count:dim_tmp+count-1] = covar_set
        count +=dim_tmp
    end
    return cov_tot
end
##


ydat= mass_cat[1][2].obs
test = covar_multi_id(ydat)
aa = inv(test)
aa = (test + test')/2
issymmetric(aa)
isposdef(aa)

evals = Diagonal(test)^(-0.5)
bb = evals * test * evals
##
imshow(bb, cmap="hot")
title("covariance")
colorbar()
display(gcf())
close("all")
##

dim = sum(size(test[i],1) for i in 1:length(test) )
cov_tot = zeros(dim, dim)
count = 1
for (i, covar) in enumerate(test)
    dim_tmp = size(test[i],1)
    println(dim_tmp)
    cov_tot[count:dim_tmp+count-1, count:dim_tmp+count-1] = covar
    count+=dim_tmp
end


## PARAMETER P[4] id FD and FDs fits
0.036 +/- 0.11
0.038 +/- 0.117
0.049 +/- 0.117
0.055 +/- 0.11
0.049 +/- 0.11
0.410 +/- 0.25
0.051 +/- 0.11
0.051 +/- 0.11
0.050 +/- 0.11
0.051 +/- 0.11
0.052 +/- 0.11
0.053 +/- 0.11
0.442 +/- 0.25
0.433 +/- 0.25
0.414 +/- 0.25
0.372 +/- 0.25
0.051 +/- 0.11
0.053 +/- 0.11
0.052 +/- 0.11
0.052 +/- 0.11
0.401 +/- 0.25
0.401 +/- 0.25
0.401 +/- 0.25
0.404 +/- 0.25
0.405 +/- 0.25
0.406 +/- 0.25
0.052 +/- 0.117
0.401 +/- 0.25
0.407 +/- 0.251
0.406 +/- 0.25
0.405 +/- 0.25
0.405 +/- 0.25