#################################
# This file was created by Alessandro Conigli - 2023
# This code is equivalent to gevp_tm.jl, but here plateaus are not passed from
# an external file, rather they are autamatically computed using a Bayesian Model Average.
# The pourpose of this code is to compute twisted mass 
# observables from raw correlators stored in mesons.dat files.
# The analysis is performed at the level of a single ensemble 
# and results are then stored in BDIO files 
# It extracts:
#  - charm quark mass 
#  - ll hl hh pseudo and vector masses 
#  - ll hl hh pseudo and vector leptonic decays
#  
#  The analysis is performed by solving a GEVP 
##################################
using Base: String
using Base: @kwdef
using LaTeXStrings: length
using OrderedCollections
using juobs, BDIO, DelimitedFiles, ADerrors, LaTeXStrings, PyPlot
using ProgressMeter
#============= SET UP VARIABLES ===========#
rcParams = PyPlot.PyDict(PyPlot.matplotlib."rcParams")
rcParams["text.usetex"] =  true
rcParams["mathtext.fontset"]  = "cm"
rcParams["font.size"] =10
rcParams["axes.labelsize"] =22
rcParams["axes.titlesize"] = 18

plt.rc("text", usetex=true) # set to true once you install latex

#============ PATHS & INFO =================#
# path to data 
const path_data_w  = "/Users/alessandroconigli/Lattice/data/aux_obs_data/wilson" # required for t0 at finite a
const path_data_tm = "/Users/alessandroconigli/Lattice/data/charm_full_Dirac"    # charm correlators 
# path to plateau
const path_plat_w       = "/Users/alessandroconigli/Lattice/data/aux_obs_data/plat_wilson.txt"
const path_plat_tm_mps  = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_mps.txt"
const path_plat_tm_mvec = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_mvec.txt"
const path_plat_tm_dps  = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_dps.txt"
const path_plat_tm_dvec = "/Users/alessandroconigli/Lattice/gevp-automated-analysis/plat/plat_tm_dvec.txt"
# path to aux obs 
const path_rw = "/Users/alessandroconigli/Lattice/data/aux_obs_data/rwf"  # reweighting factors 
const path_md = "/Users/alessandroconigli/Lattice/data/aux_obs_data/md/"   # pbp dat file for mass shift 
# path to bdio
# const path_bdio_md         = "/Users/alessandroconigli/Lattice/data/bdio_charm/md/LAT22/" # path to delta m computed by me 
const path_bdio_md         = "/Users/alessandroconigli/Lattice/data/bdio_charm/AS_ligth_obs/" # path to delta m computed by AS
const path_bdio_tm         = "/Users/alessandroconigli/Lattice/data/bdio_charm/tm" # path where to store results 
const path_bdio_tm_shifted = "/Users/alessandroconigli/Lattice/data/bdio_charm/tm_shifted" # path were to store shifted results
const path_plot = "/Users/alessandroconigli/MyDrive/phd/analysis/charm_gevp_2022/CharmPaper/plots" # path were to save plots

# ensembles to analyse 
const ens_list = [ "N203"]
# reweighting and mass shift flags
const rwf = true
# println("Warning! mass_shift set to false ")
const mass_shift = false 
const tau = 3    # gevp shift parameter
const T0  = 2
const TSM = false # set whether TSM is used or not 

#@warning("\nTSM FLAG MODE: $(TSM)\n")

#=============== INCLUDES ===============#

include("types.jl")
include("tools.jl")
include("const.jl")
include("read_bdio.jl")

#=============== ENSEMBLE INFO FROM DATABASE ==================#
ensinfo = Vector{EnsInfo}(undef, length(ens_list))
for i in 1:length(ens_list)
    ens = ens_list[i]
    try
        ensinfo[i]= EnsInfo(ens, ens_db[ens], trunc_db[ens])
    catch
        error("The ensemble id ", ens, " was not found in the const.jl ens_db database.
        Please check the ensemble id or update the database")
    end
end
wpmm = Dict{String, Vector{Float64}}()
wpmm["H101"] = [-1.0, 2.0, -1.0, -1.0]  #14.0*2.86]
wpmm["H102r002"] = [-1.0, 2.0, -1.0, -1.0] #14.0*2.86]
wpmm["H105r002"] = [-1.0, 2.0, -1.0, -1.0]# 14.0*2.86]
wpmm["H400"] = [-1.0, 2.0, -1.0,  -1.0] #14.0*3.659]
wpmm["N202"] = [-1.0, 2.0, -1.0,  -1.0] #14.0*5.164]
wpmm["N200"] = [-1.0, 2.0, -1.0, -1.0]  #14.0*5.164]
wpmm["N203"] = [-1.0, 2.0, -1.0, -1.0] # 14.0*5.164]
#wpmm["D200"] = [-1.0, 4.0, -1.0, 14.0*5.164]
wpmm["D200"] = [-1.0, 2.0, -1.0, -1.0]
wpmm["N300r002"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]
wpmm["N302"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]
wpmm["J303"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]
wpmm["J500"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]
wpmm["J501"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]
wpmm["E250"] = [-1.0, 2.0, -1.0, -1.0 ] #14.0*8.595]

## TESTING TSM WITH MULTIPLE replicas
data_sl   = read_data_sloppy(path_data_tm, "E300", "G5", "G5", legacy=false)
data_corr = read_data_correction(path_data_tm, "E300", "G5", "G5", legacy=false)

vcfg_sl      = getfield.(data_sl[1], :vcfg)
replica_sl   = Int64.(maximum.(vcfg_sl))
vcfg_corr    = getfield.(data_corr[1], :vcfg)
replica_corr = Int64.(maximum.(vcfg_corr))

corr_pp, pp_tmW =  get_corr(path_data_tm, ensinfo[1], "G5", "G5", rw=false, path_rw=path_rw, tsm=TSM)

##
#=============== LOAD DATA AND STORE t0 ===============#
corr_pp  = Vector{Vector{juobs.Corr}}(undef, length(ens_list))
corr_pa  = Vector{Vector{juobs.Corr}}(undef, length(ens_list))
corr_aa  = Vector{Vector{juobs.Corr}}(undef, length(ens_list))
yw_store = Vector{Matrix{uwreal}}(undef, length(ens_list))          # store YW info from t0
w_store  = Vector{uwreal}(undef, length(ens_list))                  # store W info from t0
pp_tmW   = Vector{Vector{Vector{uwreal}}}(undef, length(ens_list))  # store W info from corr_pp
gen_mulist = Vector{Vector{Vector{Float64}}}(undef, length(ens_list)) # store mu values 

# obs storage  
obs_tm = Vector(undef, length(ensinfo))


@time begin
    @showprogress for (k,ens) in enumerate(ensinfo)
        println("\n Ensemble: ", ens.id)
        # pp corr
        corr_pp[k], pp_tmW[k], _ =  get_corr(path_data_tm, ens, "G5", "G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        # t0 
        t0_ens, yw_store[k],  w_store[k] = get_t0(path_data_w, ens, rw=rwf, path_rw=path_rw, pl=true)
        obs_tm[k] = OrderedDict()
        obs_tm[k]["t0"] = t0_ens
        # correlators
        
        a1a1, _    = get_corr(path_data_tm, ens, "G1G5", "G1G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        a2a2, _    = get_corr(path_data_tm, ens, "G2G5", "G2G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        a3a3, _    = get_corr(path_data_tm, ens, "G3G5", "G3G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        aa_tmp     = (getfield.(a1a1, :obs) .+ getfield.(a2a2, :obs) .+ getfield.(a3a3, :obs)) ./ 3
        corr_aa[k] = [juobs.Corr(aa_tmp[l], a1a1[l].kappa, a1a1[l].mu, ["G123G123", "G123G123"], a1a1[l].y0, a1a1[l].theta1, a1a1[l].theta2 ) for l in eachindex(a1a1)] 
        
        pa1, _     = get_corr(path_data_tm, ens, "G5", "G1G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        pa2, _     = get_corr(path_data_tm, ens, "G5", "G2G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        pa3, _     = get_corr(path_data_tm, ens, "G5", "G3G5", rw=rwf, path_rw=path_rw, tsm=TSM)
        pa_tmp     = (getfield.(pa1, :obs) .+ getfield.(pa2, :obs) .+ getfield.(pa3, :obs)) ./ 3
        corr_pa[k] = [juobs.Corr(pa_tmp[l], pa1[l].kappa, pa1[l].mu, ["G0G123", "G0G123"], pa1[l].y0, pa1[l].theta1, pa1[l].theta2 ) for l in eachindex(pa1)] 
   
        gen_mulist[k] = getfield.(corr_pp[k], :mu)
    end
end

## plots for paper of D, D_s and eta_c correlators 
yyd    = corr_pp[1][4].obs; uwerr.(yyd)
yyds   = corr_pp[1][5].obs; uwerr.(yyds) 
yyeta  = corr_pp[1][6].obs; uwerr.(yyeta)
xx = collect(1:length(yyd))

errorbar(xx, value.(yyd), err.(yyd).*50, fmt="s", color="forestgreen", capsize=2, mfc="none", label=L"$D$")
errorbar(xx, value.(yyds), err.(yyds).*50, fmt="s", color="tomato", capsize=2, mfc="none", label=L"$D_s$")
errorbar(xx, value.(yyeta), err.(yyeta).*50, fmt="s", color="royalblue", capsize=2, mfc="none", label=L"$\eta_c$")
yscale("log")
legend()
xlim(65,128)
xlabel(L"$t/a$")
ylabel(L"$f_{\mathrm{PP}(x_0, y_0)}$")
#  ylim(1e-22, 1e-4)
tight_layout()
display(gcf())
savefig("/Users/alessandroconigli/Desktop/corr_comparison.pdf")
close("all")
##
# test masses D200
plat = [[85,105], [85,105], [85,105], [85,90], [88,98], [90,94], [85,90], [88,94], [90,94]  ]
# test masses J501 
plat = [[145,165], [120,155], [120,155], [125,155], [145,155], [145,165], [125,150], [145,165], [145,165]  ]
# test masses E250
plat = [[115,145], [115,160], [115,160], [118,125], [125,142], [130,160], [118,125], [125,142], [130,160]  ]
# test masses J500 
plat = [[130,160],  [140,160], [138,150], [140,160], [138,150]]

mtest = meff(corr_pp[1][5], [138,150], pl=true, wpm=wpmm) 
##
#=============== ANALYSIS ===============#
Mtotps = Vector{Vector{uwreal}}(undef, length(ensinfo)) # required for decay constants
Mtotvec = Vector{Vector{uwreal}}(undef, length(ensinfo)) # required for decay constants

@time begin
    @showprogress for (k,ens) in enumerate(ensinfo)
        # gevp matrix for ps and vec masses
        # mat_mass = comp_mat_multigamma(ens, corr_pp[k], corr_pa[k], corr_aa[k])
        # m_tm_ps, m_tm_vec = gevp_mass_BMA(mat_mass, tt0=T0, pl=true, wpm=wpmm, path_plt=path_plot)
        m_tm_ps = meff.(corr_pp[k], plat, pl=true, wpm=wpmm)
        m_tm_vec = meff.(corr_aa[k], plat, pl=true, wpm=wpmm)
        # push!(m_tm_vec, uwreal([1.029, 0.001],"mhvec_D200"))

        Mtotps[k]  = m_tm_ps
        Mtotvec[k] = m_tm_vec

        mu_list = gen_mulist[k]
        mul, mus, muh = get_mu(mu_list, ens.deg)

        # split masses by sector
        mll_tm = get_ll(mu_list, m_tm_ps, ens.deg)
        mls_tm = ens.deg ? mll_tm : get_ls(mu_list, m_tm_ps, ens.deg)  # ls
        mlh_tm = get_lh(mu_list, m_tm_ps, ens.deg)                     # lh
        msh_tm = ens.deg ? mlh_tm : get_sh(mu_list, m_tm_ps, ens.deg)  # sh
        mhh_tm = get_hh(mu_list, m_tm_ps, ens.deg)  

        mlh_star_tm = get_lh(mu_list, m_tm_vec, ens.deg)
        msh_star_tm = ens.deg ? mlh_star_tm : get_sh(mu_list, m_tm_vec, ens.deg)
        mhh_star_tm = get_hh(mu_list, m_tm_vec, ens.deg)

        obs_tm[k]["mll"] = mll_tm * sqrt(8*obs_tm[k]["t0"]) 
        obs_tm[k]["mls"] = mls_tm * sqrt(8*obs_tm[k]["t0"])
        obs_tm[k]["mlh"] = mlh_tm * sqrt(8*obs_tm[k]["t0"])
        obs_tm[k]["msh"] = msh_tm * sqrt(8*obs_tm[k]["t0"])
        obs_tm[k]["mhh"] = mhh_tm * sqrt(8*obs_tm[k]["t0"])
        #required to be consistent with storage ordering
        obs_tm[k]["flh"] = 0.0
        obs_tm[k]["fsh"] = 0.0
        obs_tm[k]["fhh"] = 0.0

        obs_tm[k]["mlh_star"] = mlh_star_tm * sqrt(8*obs_tm[k]["t0"]) 
        obs_tm[k]["msh_star"] = msh_star_tm * sqrt(8*obs_tm[k]["t0"])
        obs_tm[k]["mhh_star"] = mhh_star_tm * sqrt(8*obs_tm[k]["t0"])

        #required to be consistent with storage ordering
        obs_tm[k]["flh_star"] = 0.0
        obs_tm[k]["fsh_star"] = 0.0
        obs_tm[k]["fhh_star"] = 0.0
    
        obs_tm[k]["muh"] = muh * sqrt(8*obs_tm[k]["t0"])
        obs_tm[k]["fpik"]      = 0.0
    end
end
##
# test D200 
plat = [[85,105], [85,105], [85,105], [85,90], [88,98], [90,94], [85,90], [88,94], [90,94]  ]
# test  J501 
plat = [[120,155], [120,155], [120,155], [125,155], [130,155], [132,145], [125,150], [130,145], [132,145]  ]
# test  E250 
plat = [[115,150], [120,155], [120,155], [115,128], [130,155], [135,160], [115,128], [130,145], [135,160]  ]
# test masses J500 
plat = [[130,160],  [140,160], [138,150], [140,160], [138,150]  ]

dec_const_pcvc(corr_pp[1][5],  plat[5], Mtotps[1][5], pl=true)

##
# Decay constants
close("all")
for (k,ens) in enumerate(ensinfo)
    # gevp matrices for GEVP
    # mat_dec_ps  = comp_mat_tau_shift(ens, corr_pp[k], tau)
    # mat_dec_vev = comp_mat_tau_shift(ens, corr_aa[k], tau)

    # dec_ps = gevp_dec_BMA(mat_dec_ps, Mtotps[k], t0=T0, wpm=wpmm, pl=true, n=1, pseudo=true, wilson=false )
    # dec_ps_test = get_f_tm(corr_pp[k], Mtotps[k], ens, path_plat_tm_dps, pl=true, wpm=wpmm )

    # commented for testing D200
    # path_plt_ps = joinpath(path_plot, "plateaus", ens.id, "fps")
    # dec_tm_ps = get_f_tm_BMA(corr_pp[k], Mtotps[k], ens, pl=true, wpm=wpmm, path_plt=path_plot, ps="fps")
#    
    # path_plt_vec = joinpath(path_plot, "plateaus", ens.id, "fps")
    # dec_tm_vec   = get_f_tm_BMA(corr_aa[k], Mtotvec[k], ens, pl=true, wpm=wpmm, path_plt=path_plot, ps="fvec")
    dec_tm_ps = dec_const_pcvc.(corr_pp[k], plat, Mtotps[k], pl=true, wpm=wpmm)
    dec_tm_vec = dec_const.(corr_aa[k], plat, Mtotvec[k], pl=true, wpm=wpmm)

    mu_list = gen_mulist[k]
    mul, mus, muh = get_mu(mu_list, ens.deg)

    # split masses by sector
    flh_tm = get_lh(mu_list, dec_tm_ps, ens.deg)                     # lh
    fsh_tm = ens.deg ? flh_tm : get_sh(mu_list, dec_tm_ps, ens.deg)  # sh
    fhh_tm = get_hh(mu_list, dec_tm_ps, ens.deg)                     # hh

    flh_star_tm = get_lh(mu_list, dec_tm_vec, ens.deg)
    fsh_star_tm = ens.deg ? flh_star_tm : get_sh(mu_list, dec_tm_vec, ens.deg)
    fhh_star_tm = get_hh(mu_list, dec_tm_vec, ens.deg)

    obs_tm[k]["flh"] = flh_tm * sqrt(8*obs_tm[k]["t0"])
    obs_tm[k]["fsh"] = fsh_tm * sqrt(8*obs_tm[k]["t0"])
    obs_tm[k]["fhh"] = fhh_tm * sqrt(8*obs_tm[k]["t0"])

    obs_tm[k]["flh_star"] = flh_star_tm * sqrt(8*obs_tm[k]["t0"])
    obs_tm[k]["fsh_star"] = fsh_star_tm * sqrt(8*obs_tm[k]["t0"])
    obs_tm[k]["fhh_star"] = fhh_star_tm * sqrt(8*obs_tm[k]["t0"])
end


## mass shift

## careful with H105r002! Do not substitute pion and kaon masses in this case 
if mass_shift
    
    deltam = Vector(undef, length(ensinfo))
    obs_tm_shifted = Vector(undef, length(ensinfo))
    @info "Delta m taken from $(path_bdio_md)"
    
    for (k,ens) in enumerate(ensinfo)
        # read deltam from bdio
        # read dm computed by AS
        rep = filter(x->occursin(ens.id, x), readdir(path_bdio_md,join=true))[1]
        ens.deg ? Nmolt =3. : Nmolt = 1. # required for symmetric point ensembles if shifting s only 
        deltam[k] = read_BDIO(rep, "ll_obs", "dm")[1] * Nmolt
        # these two lines for dm computed locally by me
        # fname = joinpath(path_bdio_md, string(ens.id,".bdio"))
        # deltam[k] = read_BDIO(fname, "md", "deltam")[1]
        # read dS/dm
        dSdm = read_pbp(path_md, ens.id)
        # definitions
        prim_obs = vcat(pp_tmW[k]...)
        obs_tm_shifted[k] = deepcopy(obs_tm[k])
        
        ## LOOOP OVER OBSERVABLES + MASS SHIFT
        for (key, value) in obs_tm[k] 
            if isa(value, uwreal) # if 1 uwreal
                mds1, mds2 = md_sea(value, dSdm, prim_obs, w_store[k]) .+  md_sea(value, dSdm, yw_store[k], w_store[k])
                # obs_tm_shifted[k][key] = obs_tm[k][key] + deltam[k] * (2 * mds1 + mds2) # shifting u,d,s
                obs_tm_shifted[k][key] = obs_tm[k][key] + deltam[k] * (mds2) # shifting s
            elseif isa(value, Vector{uwreal}) # if vector uwreal
                for (i, v) in enumerate(value)
                    mds1, mds2 =  md_sea(v, dSdm, prim_obs, w_store[k]) .+  md_sea(v, dSdm, yw_store[k], w_store[k])
                    # obs_tm_shifted[k][key][i] = obs_tm[k][key][i] + deltam[k] * (2 * mds1 + mds2) # shifting u,d,s
                    obs_tm_shifted[k][key][i] = obs_tm[k][key][i] + deltam[k] * (mds2) # shifting s
                end
            end
        end
    end
end

## read and store fpik and mpi from AS analysis 
for (k,ens) in enumerate(ensinfo)
    rep = filter(x->occursin(ens.id, x), readdir(path_bdio_md,join=true))[1]
    t0_aux  = read_BDIO(rep, "ll_obs", "t0")[1] # check that your t0_shifted coincide with this
    mpi_aux = read_BDIO(rep, "ll_obs", "mpi")[1]
    mk_aux = read_BDIO(rep, "ll_obs", "mk")[1]
    fpi_aux = read_BDIO(rep, "ll_obs", "fpi")[1]
    fk_aux  = read_BDIO(rep, "ll_obs", "fk")[1]

    sqrt8t0 = sqrt(8*t0_aux)     # sqrt(8*obs_tm_shifted[k]["t0"])
    obs_tm_shifted[k]["mll"] =  sqrt8t0 * mpi_aux 
    obs_tm_shifted[k]["mls"] =  sqrt8t0 * mk_aux 

    obs_tm_shifted[k]["fpik"] = sqrt8t0 * (2. / 3.) * (fk_aux + 0.5*fpi_aux)
    obs_tm[k]["fpik"] = sqrt8t0 * (2. / 3.) * (fk_aux + 0.5*fpi_aux)

end

##

[pop!(obs_tm[k], "fpk") for k in 1:length(obs_tm)]
[pop!(obs_tm_shifted[k], "fpk") for k in 1:length(obs_tm)]

##
#====================== WRITE BDIO =====================#
[pop!(obs_tm[k], "t0") for k in 1:length(obs_tm)]
if mass_shift
    #TM SHIFTED
    [pop!(obs_tm_shifted[k], "t0") for k in 1:length(obs_tm_shifted)]
    for (k, obs) in enumerate(obs_tm_shifted) #loop over ensembles
        ens = ensinfo[k].id
        p = joinpath(path_bdio_tm_shifted, string(ens, ".bdio")) # 
        fb = BDIO_open(p, "w", ens) 
        uinfo = 0
        for (key, value) in obs #loop over observables
            if isa(value, uwreal)
                write_uwreal(value, fb, uinfo)
            elseif isa(value, Vector{uwreal})
                [write_uwreal(v, fb, uinfo) for v in value]
            end
            uinfo += 1
        end
        BDIO_close!(fb)
    end
end
# TM
for (k, obs) in enumerate(obs_tm) #loop over ensembles
    ens = ensinfo[k].id
    p = joinpath(path_bdio_tm, string(ens, ".bdio"))
    fb = BDIO_open(p, "w", ens)
    uinfo = 0
    for (key, value) in obs #loop over observables
        if isa(value, uwreal)
            write_uwreal(value, fb, uinfo)
        elseif isa(value, Vector{uwreal})
            [write_uwreal(v, fb, uinfo) for v in value]
        end
        uinfo += 1
    end
    BDIO_close!(fb)
end
